uniform sampler2D mytex[6];
uniform sampler2D normalMap;
uniform sampler2D heightMap;
//uniform sampler2D distanceMap;
uniform float levelRatio;
uniform float slopeCutoff;
uniform float triplanarCutoff;
uniform bool lighting;
uniform bool texturing;
uniform bool slopeTexturing;
uniform bool normalTexture;
uniform bool useDetailMap;
uniform bool triplanarTexturing;
uniform bool textureLOD;
uniform bool useTexCorrection;
uniform int frame;
uniform float texRep;
uniform float waterPosition;
varying vec3 lightDir;
varying float distance;
varying float smoothLod;

varying float test;

varying float lodLevel;

#define LODS 7

#define INSANE_DETAIL 7
#define EXTREME_DETAIL 6
#define REALLY_HIGH_DETAIL 5
#define HIGH_DETAIL 4
#define MEDIUM_HIGH_DETAIL 3
#define MEDIUM_DETAIL 2
#define LOW_DETAIL 1

const int noLevs = 5;
float[7] lodReps = float[](18.0, 6.0, 3.5, 2.0, 1.25, 0.75, 0.25);
float[7] lodCutoffs = float[](0.0, 30.0, 60.0, 120.0, 240.0, 480.0, 960.0);

vec2 texDistance = vec2(0.0,0.0);

// Find the terrain level based on height
// Warning: multiple returns
int findTerrainLevel(float height)
{
	if (height <= 0.001) return 0;
	for (int i = 1; i < noLevs; i++) //multiple returns, loop may break early
	{
		if (height <= levelRatio * i) return i;
	}
	return noLevs;
}

int getLevelOfDetail()
{
	if (distance < lodCutoffs[1]) return INSANE_DETAIL;
	else if (distance < lodCutoffs[2]) return EXTREME_DETAIL;
	else if (distance < lodCutoffs[3]) return REALLY_HIGH_DETAIL;
	else if (distance < lodCutoffs[4]) return HIGH_DETAIL;
	else if (distance < lodCutoffs[5]) return MEDIUM_HIGH_DETAIL;
	else if (distance < lodCutoffs[6]) return MEDIUM_DETAIL;
	else return LOW_DETAIL;
}

int round(float num)
{
	float result = floor(num);
	float res1 = num - result;
	if (res1 >= 0.5) result = ceil(num);
	return result;
}

// Should be used for polygons which cross LOD boundaries
// to avoid distortions due to sudden repetition change.
vec4 getLevelOfDetailColour(sampler2D tex, float coordMod)
{
	vec4 colour1;
	vec4 colour2;
	float thisSmoothLod = smoothLod;

	if (abs(lodLevel-round(lodLevel))>0.000005)
	{
		float cut1 = 30*pow(2.0,floor(lodLevel)-2.0);
		float cut2 = cut1*2.0;
		thisSmoothLod = smoothstep(cut1,cut2,distance);
		float rep1 = 10*pow(0.5,floor(lodLevel)-2.0);
    		float rep2 = rep1/2.0;
		colour1 = texture2D(tex, (gl_TexCoord[6].st+texDistance)*rep1*texRep+coordMod);
		colour2 = texture2D(tex, (gl_TexCoord[6].st+texDistance)*rep2*texRep+coordMod);
	}
	else
	{
		colour1 = texture2D(tex, gl_TexCoord[0].st+coordMod+texDistance);
		colour2 = texture2D(tex, gl_TexCoord[7].st+coordMod+texDistance);
	}

	return mix(colour1,colour2,thisSmoothLod);
}

// Calculates lighting for this fragment given the normal and lightDirection (global)
// Returns the shade value (greyscale)
vec4 shadow(vec4 normalFromMap)
{
	vec3 normal = gl_NormalMatrix * normalFromMap.xyz;
	vec3 N = normalize(normal);
	vec3 L = normalize(lightDir);
	float lambertTerm = dot(N,L);
	vec4 shade = vec4(0.4,0.4,0.4,1.0);
	if (lambertTerm > 0.0) shade+= lambertTerm*vec4(0.6,0.6,0.6,1); 
	return shade;
}

// Applies a given colour to highly sloped areas and blends with current colour for
// "in between" slope values. Returns the resultant colour.
vec4 applySlopeTexturing(vec4 currentColour, vec4 slopeColour, float slope, int level)
{
	vec4 finalColour = currentColour;
	//don't apply slope texturing for top level (assuming heavy snow) or bottom level
	if ((level < noLevs-1) && (level > 0))
	{
		float difference = smoothstep(slopeCutoff/1.25, slopeCutoff, slope);
		finalColour = mix(slopeColour, currentColour, difference);
	}
	return finalColour;
}

// Changes the way textures are sampled based on normal direction.
// Normals with a large(ish) z component and/or x component will use a
// different combination of samples as calculated in the vertex shader.
// Sluggish. Requires many texture samples (up to 3 times the amount)
vec4[noLevs] applyTriplanarTexturing(vec4 normal, bool useDetMap, int lod)
{
	float absX = abs(normal.x);
	float absY = abs(normal.y);
	float absZ = abs(normal.z);

	vec4 textures[noLevs];
	for (int i = 0; i < noLevs; i++)
	{
		textures[i] = vec4(0,0,0,0);
	}
	vec4 detail = vec4(1,1,1,1);
	if (useDetMap) detail = vec4(0,0,0,0);

	//following values are kind of hacked via experimentation
	if ((absZ > triplanarCutoff) && (absZ > absX))
	{
		float difference;
		difference = smoothstep(triplanarCutoff,0.8,absZ);
		if (useDetMap) detail += texture2D(mytex[noLevs], gl_TexCoord[5].st)*difference;
		for (int i = 0; i < noLevs; i++)
		{
			textures[i] += texture2D(mytex[i], gl_TexCoord[2].st*lodReps[lod])*difference;
		}
	}	
	else if (absX > triplanarCutoff)
	{
		float difference;
		difference = smoothstep(triplanarCutoff,0.8,absX);
		if (useDetMap) detail += texture2D(mytex[noLevs], gl_TexCoord[4].st)*difference;
		for (int i = 0; i < noLevs; i++)
		{
			textures[i] += texture2D(mytex[i], gl_TexCoord[1].st*lodReps[lod])*difference;
		}			
	}

	float difference = 1.0-textures[0].a; //alpha should be the same for all textures at this stage
	if (useDetMap) detail += texture2D(mytex[noLevs], gl_TexCoord[3].st)*difference;
	textures[0] += texture2D(mytex[0], gl_TexCoord[0].st+waterPosition)*difference;
	if (useDetMap) detail*=2.0;
	for (int i = 1; i < 4; i++)
	{
		textures[i] += texture2D(mytex[i], gl_TexCoord[0].st*lodReps[lod])*difference*detail;
	}	
	textures[4] += texture2D(mytex[4], gl_TexCoord[0].st*lodReps[lod])*difference;

	return textures;
}

// Converts the 4-dimensional color values in the distance map into
// a 2-dimensional (x,y) pair that can be used to offset textures
/*vec2 distanceMapToDistance()
{
	vec4 distance = texture2D(distanceMap, gl_TexCoord[6].st);
	
	float x = (distance.r * 65025.0 + distance.g*255.0)/65025.0;
	float y = (distance.b * 65025.0 + distance.a*255.0)/65025.0;

	return vec2(x*texRep,y*texRep);
}*/

// Method: Calculate texture colours for every texture (!!) which includes any detail
// and any slope based texture sampling. Find the current height level and apply the
// appropriate colour/s, blending as necessary. Apply a slope based texture and 
// finally calculate and apply lighting.
//
// Possible improvement: Only calculate texture samples that are required for the 
// current height level.
void main()
{
	float difference;
	int lod = getLevelOfDetail();

	bool useDetMap = ((useDetailMap) && (lod >= HIGH_DETAIL));

	vec4 height = texture2D(heightMap, gl_TexCoord[6].st);
	vec4 normals = texture2D(normalMap, gl_TexCoord[6].st);
	
	//if (useTexCorrection) texDistance = distanceMapToDistance();

	float h = height.a;
	int level = findTerrainLevel(h);

	normals = normals * 2.0 - 1.0; //changing [0, 1] to [-1, 1]

	vec4 textures[noLevs];
	vec4 detail = vec4(1,1,1,1);
	if ((triplanarTexturing)&&(lod>=MEDIUM_DETAIL)) 
		textures = applyTriplanarTexturing(normals,useDetMap, LODS-lod);
	else
	{	
		if (useDetMap)
			detail = texture2D(mytex[noLevs], gl_TexCoord[3].st)*2.0;
		//animated water
		//textures[0] = texture2D(mytex[0], gl_TexCoord[0].st+waterPosition);
		textures[0] = getLevelOfDetailColour(mytex[0],waterPosition);
		for (int i = 1; i < noLevs-1; i++)
		{
			if (textureLOD) 
			textures[i] = getLevelOfDetailColour(mytex[i],0.0)*detail;
			else 
			textures[i] = texture2D(mytex[i], gl_TexCoord[0].st+texDistance)*detail;
		}
		if (textureLOD) 
		textures[noLevs-1]=getLevelOfDetailColour(mytex[noLevs-1],0.0);
		else 
		textures[noLevs-1] = texture2D(mytex[noLevs-1],gl_TexCoord[0].st+texDistance);
		
	}

	vec4 colour = vec4(1,1,1,1);
	vec4 final_colour;
	
	//apply the appropriate colour textures
	//difference is the ratio of a point's height within its
	//height level
	if (texturing)
	{
		if (level == 0) colour = textures[0];
		else if (level == 1) colour = textures[1];
		else if (level < noLevs)
		{
			//difference = smoothstep(levelRatio*(level-1.0), levelRatio*level, h);
			difference = smoothstep(levelRatio*(level-0.75), levelRatio*(level-0.25), h);
			colour = mix(textures[level-1], textures[level], difference);
		}
		else colour = textures[level-1];
	}

	if (slopeTexturing)
		final_colour = applySlopeTexturing(colour,textures[3],normals.y, level);
	else
		final_colour = colour;

	if (lighting) final_colour *= shadow(normals);

	if (normalTexture) final_colour = normals;

	gl_FragColor = final_colour;
	//TESTING
	//gl_FragColor = vec4(texture2D(distanceMap, gl_TexCoord[6].st).r,0,0,0);
	//gl_FragColor = texture2D(distanceMap, gl_TexCoord[6].st);
	//_FragColor = vec4(texDistance,0,0);
}